/*******************************************************************************
*    TimeAdjust 1.0 for PalmOS                                                 *
*    Copyright (C) 2009 Luca Bertoncello                                       *
*    Hartigstrasse, 12 - 01127 Dresden Deutschland                             *
*    E-Mail: lucabert@lucabert.de, lucabert@lucabert.com                       *
*    http://www.lucabert.de/  http://www.lucabert.com/                         *
*                                                                              *
*    This program is free software; you can redistribute it and/or modify      *
*    it under the terms of the GNU General Public License as published by      *
*    the Free Software Foundation; version 2 of the License.                   *
*                                                                              *
*    This program is distributed in the hope that it will be useful,           *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of            *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
*    GNU General Public License for more details.                              *
*                                                                              *
*    You should have received a copy of the GNU General Public License         *
*    along with this program; if not, write to the Free Software               *
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
*******************************************************************************/

#include <PalmOS.h>
#include <PalmTypes.h>
#include <PalmCompatibility.h>
#include <AlarmMgr.h>
#include <TimeMgr.h>
#include <DateTime.h>
#include <LocaleMgr.h>
#include <System/Unix/unix_stdio.h>
#include <System/Unix/sys_types.h>
#include <System/Unix/sys_socket.h>
#include <System/Unix/sys_time.h>
#include <System/Unix/netinet_in.h>
#include <System/Unix/arpa_inet.h>
#include <System/Unix/netdb.h>
#include "timeAdjust.h"
#include "prefs.h"

#define appID                   'TAd'
#define versionID               1
#define NTP_PORT                123
#define PILOTOFFSET             126144000L

struct {
    unsigned char livnmode, stratum, poll, prec;
    unsigned long delay, disp, refid;
    unsigned long reftime[2], origtime[2], rcvtime[2], txtime[2];
} pkt;

int netisopen;
Word AppNetRefnum;
Long AppNetTimeout;
Err errno, h_errno;
NetHostInfoBufType AppHostInfo;
NetServInfoBufType AppServInfo;
UInt32 newTime;

static Err OpenNet(void)
{
  Err err;
  Word ifErrs;

  if(netisopen)
      return 0;

  err = SysLibFind("Net.lib", &AppNetRefnum);
  if (err)
    return err;

  err = NetLibOpen(AppNetRefnum, &ifErrs);
  if((err && err != netErrAlreadyOpen) || ifErrs)
    return err;

  AppNetTimeout = SysTicksPerSecond() * 4;
  netisopen++;

  return 0;
}

static Err CloseNet(void)
{
  if(!netisopen)
    return 0;

  NetLibClose(AppNetRefnum, false);
  netisopen = 0;
  return 0;
}

static Err StartApplication() 
{
  FrmGotoForm(MainForm);
  return 0;
}

static void StopApplication() 
{
  FrmCloseAllForms();
}

Char* GetField(UInt16 formID, UInt16 fieldID)
{
  FormPtr frm;
  FieldPtr fld;
  Char *p;

  frm = FrmGetFormPtr(formID);
  fld = (FieldPtr)FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  p = FldGetTextPtr(fld);

  return (p ? p : "");
}

void SetField(UInt16 formID, UInt16 fieldID, MemPtr str)
{
  FormPtr frm;
  FieldPtr fld;
  CharPtr p;
  VoidHand h;

  frm = FrmGetFormPtr(formID);
  fld = (FieldPtr) FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, fieldID));
  h = (VoidHand) FldGetTextHandle(fld);
  if(h == NULL)
  {
    h = MemHandleNew(FldGetMaxChars(fld)+1);
    ErrFatalDisplayIf(!h, "No Memory");
  }

  p = (CharPtr) MemHandleLock(h);
  StrCopy(p, str);
  MemHandleUnlock(h);
  FldSetTextHandle(fld, (Handle) h);
}

void displayTimeCorrections(UInt32 lastCorrection, UInt32 nextCorrection)
{
  DateTimeType dateTimeL, dateTimeN;
  MemHandle H;
  Char buf[255], strL[255], strN[255];
  FormPtr frm = FrmGetFormPtr(MainForm);

  TimSecondsToDateTime(lastCorrection, &dateTimeL);
  TimSecondsToDateTime(nextCorrection, &dateTimeN);
  H = DmGetResource(strRsc, dateTimeFmt);
  StrCopy(buf, (char*)MemHandleLock(H));

  if(strcmp(buf, "DE") == 0)
  {
    if(lastCorrection != 0)
      sprintf(strL, "%02d.%02d.%4d %02d:%02d.%02d", dateTimeL.day, dateTimeL.month, dateTimeL.year, dateTimeL.hour, dateTimeL.minute, dateTimeL.second);
    else
      strcpy(strL, "--");
    sprintf(strN, "%02d.%02d.%4d %02d:%02d.%02d", dateTimeN.day, dateTimeN.month, dateTimeN.year, dateTimeN.hour, dateTimeN.minute, dateTimeN.second);
  }
  else if(strcmp(buf, "EN") == 0)
  {
    if(lastCorrection != 0)
      sprintf(strL, "%02d/%02d/%4d %02d:%02d.%02d", dateTimeL.month, dateTimeL.day, dateTimeL.year, dateTimeL.hour, dateTimeL.minute, dateTimeL.second);
    else
      strcpy(strL, "--");
    sprintf(strN, "%02d/%02d/%4d %02d:%02d.%02d", dateTimeN.month, dateTimeN.day, dateTimeN.year, dateTimeN.hour, dateTimeN.minute, dateTimeN.second);
  }
  else if(strcmp(buf, "IT") == 0)
  {
    if(lastCorrection != 0)
      sprintf(strL, "%02d/%02d/%4d %02d:%02d.%02d", dateTimeL.day, dateTimeL.month, dateTimeL.year, dateTimeL.hour, dateTimeL.minute, dateTimeL.second);
    else
      strcpy(strL, "--");
    sprintf(strN, "%02d/%02d/%4d %02d:%02d.%02d", dateTimeN.day, dateTimeN.month, dateTimeN.year, dateTimeN.hour, dateTimeN.minute, dateTimeN.second);
  }

  SetField(MainForm, idLastCorrection, strL);
  SetField(MainForm, idNextCorrection, strN);
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastCorrection)));
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idNextCorrection)));
  MemHandleUnlock(H);
  DmReleaseResource(H);
}

void displayTimeSync(UInt32 lastSync)
{
  DateTimeType dateTimeS;
  MemHandle H;
  Char buf[255], strS[255];
  FormPtr frm = FrmGetFormPtr(MainForm);

  TimSecondsToDateTime(lastSync, &dateTimeS);
  H = DmGetResource(strRsc, dateTimeFmt);
  StrCopy(buf, (char*)MemHandleLock(H));

  if(strcmp(buf, "DE") == 0)
  {
    if(lastSync != 0)
      sprintf(strS, "%02d.%02d.%4d %02d:%02d.%02d", dateTimeS.day, dateTimeS.month, dateTimeS.year, dateTimeS.hour, dateTimeS.minute, dateTimeS.second);
    else
      strcpy(strS, "--");
  }
  else if(strcmp(buf, "EN") == 0)
  {
    if(lastSync != 0)
      sprintf(strS, "%02d/%02d/%4d %02d:%02d.%02d", dateTimeS.month, dateTimeS.day, dateTimeS.year, dateTimeS.hour, dateTimeS.minute, dateTimeS.second);
    else
      strcpy(strS, "--");
  }
  else if(strcmp(buf, "IT") == 0)
  {
    if(lastSync != 0)
      sprintf(strS, "%02d/%02d/%4d %02d:%02d.%02d", dateTimeS.day, dateTimeS.month, dateTimeS.year, dateTimeS.hour, dateTimeS.minute, dateTimeS.second);
    else
      strcpy(strS, "--");
  }

  SetField(MainForm, idLastSync, strS);
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
  MemHandleUnlock(H);
  DmReleaseResource(H);
}

void setNextCorrectionAlarm(UInt32 nextCorrection)
{
  AlmSetProcAlarm((AlmAlarmProcPtr) adjustTime, 0, nextCorrection);
}

void adjustTime(Word almProcCmd, SysAlarmTriggeredParamType* cmd)
{
  UInt16 nAppPrefsSize;
  Preferences gAppPrefs;
  UInt32 drift;

  cmd->purgeAlarm = true;

  nAppPrefsSize = sizeof(Preferences);
  if(PrefGetAppPreferences(appID, 1, &gAppPrefs, &nAppPrefsSize, true) == noPreferenceFound)
    MemSet(&gAppPrefs, sizeof(Preferences), 0);
  if(gAppPrefs.secDay < 0)
  {
    TimSetSeconds(TimGetSeconds() + 1);
    drift = -86400 / gAppPrefs.secDay;
  }
  else
  {
    TimSetSeconds(TimGetSeconds() - 1);
    drift = 86400 / gAppPrefs.secDay;
  }
  gAppPrefs.lastCorrection = TimGetSeconds();
  gAppPrefs.nextCorrection = TimGetSeconds() + drift;
  if((gAppPrefs.nextCorrection % 60) < 5)
    gAppPrefs.nextCorrection += 10;
  PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);

  setNextCorrectionAlarm(gAppPrefs.nextCorrection);
}

int sntpGetTime(char *ntpServer)
{
  int fd = -1, len, n;
  struct sockaddr_in saddr;
  struct hostent *hp;
  FormPtr frm = FrmGetFormPtr(MainForm);
  MemHandle H;
  Char buf[255];

  Int16 timeZone = PrefGetPreference(prefTimeZone); 
  Int16 daylightSavingAdjustment = PrefGetPreference(prefDaylightSavingAdjustment);

  if(OpenNet())
    return -1;

  H = DmGetResource(strRsc, ntpResolveHost);
  StrCopy(buf, (char*)MemHandleLock(H));
  SetField(MainForm, idLastSync, buf);
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
  MemHandleUnlock(H);
  DmReleaseResource(H);

  bzero(&saddr, sizeof saddr);
  saddr.sin_family = AF_INET;
  saddr.sin_port = htons(NTP_PORT);
  if((saddr.sin_addr.s_addr = inet_addr(ntpServer)) == -1)
  {
    if ((hp = gethostbyname(ntpServer)) == NULL)
      return -1;
    bcopy(hp->h_addr, &saddr.sin_addr, sizeof saddr.sin_addr);
  }

  H = DmGetResource(strRsc, ntpConnect);
  StrCopy(buf, (char*)MemHandleLock(H));
  SetField(MainForm, idLastSync, buf);
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
  MemHandleUnlock(H);
  DmReleaseResource(H);

  if((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
  {
    CloseNet();
    H = DmGetResource(strRsc, ntpError);
    StrCopy(buf, (char*)MemHandleLock(H));
    SetField(MainForm, idLastSync, buf);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
    MemHandleUnlock(H);
    DmReleaseResource(H);
    return -2;
  }

  if(connect(fd, (struct sockaddr *) &saddr, sizeof(struct sockaddr_in)) < 0)
  {
    close(fd);
    CloseNet();
    H = DmGetResource(strRsc, ntpError);
    StrCopy(buf, (char*)MemHandleLock(H));
    SetField(MainForm, idLastSync, buf);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
    MemHandleUnlock(H);
    DmReleaseResource(H);
    return -3;
  }

  H = DmGetResource(strRsc, ntpSynchonize);
  StrCopy(buf, (char*)MemHandleLock(H));
  SetField(MainForm, idLastSync, buf);
  FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
  MemHandleUnlock(H);
  DmReleaseResource(H);

  bzero(&pkt, sizeof pkt);
  pkt.livnmode = 0x0b;
  if((n = send(fd, &pkt, sizeof pkt, 0)) != sizeof pkt)
  {
    close(fd);
    CloseNet();
    H = DmGetResource(strRsc, ntpError);
    StrCopy(buf, (char*)MemHandleLock(H));
    SetField(MainForm, idLastSync, buf);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
    MemHandleUnlock(H);
    DmReleaseResource(H);
    return -4;
  }

  len = sizeof pkt;
  if((n = recvfrom(fd, &pkt, sizeof pkt, 0, NULL, NULL)) != sizeof pkt)
  {
    close(fd);
    CloseNet();
    H = DmGetResource(strRsc, ntpError);
    StrCopy(buf, (char*)MemHandleLock(H));
    SetField(MainForm, idLastSync, buf);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
    MemHandleUnlock(H);
    DmReleaseResource(H);
    return -5;
  }

  if((pkt.livnmode & 0xc0) == 0xc0 || !pkt.stratum || !pkt.txtime[0])
  {
    close(fd);
    CloseNet();
    H = DmGetResource(strRsc, ntpError);
    StrCopy(buf, (char*)MemHandleLock(H));
    SetField(MainForm, idLastSync, buf);
    FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastSync)));
    MemHandleUnlock(H);
    DmReleaseResource(H);
    return -6;
  }

  newTime = TimUTCToTimeZone((pkt.txtime[0]) - PILOTOFFSET, timeZone, daylightSavingAdjustment);

  close(fd);
  CloseNet();

  return 0;
}

Boolean AboutFormEventHandler(EventPtr event)
{
  Boolean handled = false;
  FormPtr aboutFrm = FrmGetFormPtr(AboutForm);

  switch(event->eType)
  {
    case frmOpenEvent:
      FrmDrawForm(aboutFrm);
      handled = true;
    break;

    case frmCloseEvent:
      FrmEraseForm(aboutFrm);
      FrmDeleteForm(aboutFrm);
      handled = true;
    break;

    case ctlSelectEvent:
      switch (event->data.ctlSelect.controlID)
      {
        case idCloseAbout:
          FrmReturnToForm(MainForm);
          handled = true;
        break;
      }
    break;

    default:
    break;
  }

  return handled;
}

Boolean MainFormEventHandler(EventPtr event)
{
  Boolean handled = false;
  FormPtr frm = FrmGetFormPtr(MainForm);
  UInt16 nAppPrefsSize;
  CharPtr str;
  UInt32 drift;
  Preferences gAppPrefs;
  int ret;

  str = MemPtrNew(30);
  switch(event->eType)
  {
    case frmOpenEvent:
      FrmDrawForm(FrmGetFormPtr(MainForm));
      nAppPrefsSize = sizeof(Preferences);
      if(PrefGetAppPreferences(appID, 1, &gAppPrefs, &nAppPrefsSize, true) == noPreferenceFound)
        MemSet(&gAppPrefs, sizeof(Preferences), 0);
      CtlSetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idIsActive)), gAppPrefs.active);
      StrIToA(str, gAppPrefs.secDay);
      SetField(MainForm, idSecDay, str);
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idSecDay)));
      displayTimeSync(gAppPrefs.lastSync);
      if(strlen(gAppPrefs.ntpServer) == 0)
      {
        strcpy(gAppPrefs.ntpServer, "pool.ntp.org");
        PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);
      }
      SetField(MainForm, idNTPServer, gAppPrefs.ntpServer);
      FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idNTPServer)));
      
      if(gAppPrefs.active)
        displayTimeCorrections(gAppPrefs.lastCorrection, gAppPrefs.nextCorrection);
      else
      {
        SetField(MainForm, idLastCorrection, " -- ");
        SetField(MainForm, idNextCorrection, " -- ");
        FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastCorrection)));
        FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idNextCorrection)));
      }
      handled = true;
    break;

    case frmCloseEvent:
      FrmEraseForm(FrmGetFormPtr(MainForm));
      FrmDeleteForm(FrmGetFormPtr(MainForm));
      handled = true;
    break;

    case ctlSelectEvent:
      switch (event->data.ctlSelect.controlID)
      {
        case idSecPlus:
          StrIToA(str, StrAToI(GetField(MainForm, idSecDay)) + 1);
          SetField(MainForm, idSecDay, str);
          FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idSecDay)));
        break;
        case idSecMinus:
          StrIToA(str, StrAToI(GetField(MainForm, idSecDay)) - 1);
          SetField(MainForm, idSecDay, str);
          FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idSecDay)));
        break;
        case idButtonSync:
          nAppPrefsSize = sizeof(Preferences);
          if(PrefGetAppPreferences(appID, 1, &gAppPrefs, &nAppPrefsSize, true) == noPreferenceFound)
            MemSet(&gAppPrefs, sizeof(Preferences), 0);
          strcpy(gAppPrefs.ntpServer, GetField(MainForm, idNTPServer));

          ret = sntpGetTime(gAppPrefs.ntpServer);
          if(ret == 0)
          {
            gAppPrefs.lastSync = newTime;
            TimSetSeconds(newTime);
          }

          displayTimeSync(gAppPrefs.lastSync);
          gAppPrefs.active = CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idIsActive)));
          gAppPrefs.secDay = StrAToI(GetField(MainForm, idSecDay));
          if(gAppPrefs.active && gAppPrefs.secDay != 0)
          {
            if(gAppPrefs.secDay < 0)
              drift = -86400 / gAppPrefs.secDay;
            else if(gAppPrefs.secDay > 0)
              drift = 86400 / gAppPrefs.secDay;
            else
              drift = 0;
            gAppPrefs.nextCorrection = TimGetSeconds() + drift;
            if((gAppPrefs.nextCorrection % 60) < 5)
              gAppPrefs.nextCorrection += 10;
            displayTimeCorrections(gAppPrefs.lastCorrection, gAppPrefs.nextCorrection);

            setNextCorrectionAlarm(gAppPrefs.nextCorrection);
          }
          PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);
        break;
        case idButtonSave:
          gAppPrefs.active = CtlGetValue(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idIsActive)));
          gAppPrefs.secDay = StrAToI(GetField(MainForm, idSecDay));
          strcpy(gAppPrefs.ntpServer, GetField(MainForm, idNTPServer));
          if(gAppPrefs.active && gAppPrefs.secDay != 0)
          {
            if(gAppPrefs.secDay < 0)
              drift = -86400 / gAppPrefs.secDay;
            else if(gAppPrefs.secDay > 0)
              drift = 86400 / gAppPrefs.secDay;
            else
              drift = 0;
            gAppPrefs.nextCorrection = TimGetSeconds() + drift;
            if((gAppPrefs.nextCorrection % 60) < 5)
              gAppPrefs.nextCorrection += 10;
            displayTimeCorrections(gAppPrefs.lastCorrection, gAppPrefs.nextCorrection);

            setNextCorrectionAlarm(gAppPrefs.nextCorrection);
          }
          else
          {
            setNextCorrectionAlarm(0);
            SetField(MainForm, idLastCorrection, " -- ");
            SetField(MainForm, idNextCorrection, " -- ");
            FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idLastCorrection)));
            FldDrawField(FrmGetObjectPtr(frm, FrmGetObjectIndex(frm, idNextCorrection)));
          }
          PrefSetAppPreferences(appID, 1, versionID, &gAppPrefs, sizeof(Preferences), true);

          handled = true;
        break;
      }
    break;

    default:
    break;
  }
  MemPtrFree(str);

  return handled;
}

static Boolean ApplicationEventHandler(EventPtr event) 
{
  Boolean handled = false;
  FormType* frmP;
  ListType *lstP;

  switch(event->eType)
  {
    case frmLoadEvent:
    {
      FormPtr pForm = FrmInitForm(event->data.frmLoad.formID);
      FrmSetActiveForm(pForm);
      switch(event->data.frmLoad.formID)
      {
        case MainForm:
          FrmSetEventHandler(pForm, MainFormEventHandler);
        break;

        case AboutForm:
          FrmSetEventHandler(pForm, AboutFormEventHandler);
        break;

        default:
        break;
      }
      handled = true;
    }
    break;

    case menuEvent:
      switch(event->data.menu.itemID)
      {
      	case HelpMenuAbout:
          FrmPopupForm(AboutForm);
        break;
      	case HelpMenuHelp:
          FrmAlert(HelpAlert);
        break;
      }
      handled = true;
    break;

    default:
    break;
  }

  return handled;
}

static void EventLoop() 
{
  EventType event;
  UInt16 error;

  do
  {
    EvtGetEvent(&event, evtWaitForever);

    if(!SysHandleEvent(&event))
    {
      if(!MenuHandleEvent(0, &event, &error))
      {
        if(!ApplicationEventHandler(&event))
          FrmDispatchEvent(&event);
      }
    }
  } while(event.eType != appStopEvent);
}

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) 
{
  Err err = 0;
  UInt32 romVersion;

  switch(launchCode)
  {
    case sysAppLaunchCmdNormalLaunch:
    {
      FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
      if(romVersion < sysMakeROMVersion(3, 5, 0, sysROMStageRelease, 0))
      {
        FrmAlert(RomIncompatibleAlert);
        return sysErrRomIncompatible;
      }
      if((err = StartApplication()) == 0)
      {
        EventLoop();
        StopApplication();
      }
      break;
    }
    break;
  }

  return err;
}
